/*
 *  libxcpc by Davide Libenzi (Exception and resource handling in C)
 *  Copyright (C) 2002  Davide Libenzi
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Davide Libenzi <davidel@xmailserver.org>
 *
 */

#include "xcpc-include.h"



static char *xcpc_excpt_strdup(xcpc_ctx ctx, char const *str);
static void *xcpc_excpt_memdup(xcpc_ctx ctx, void const *data, int size);




static char *xcpc_excpt_strdup(xcpc_ctx ctx, char const *str) {

	return xcpc__strdup(xcpc_context_exhandler(ctx), NULL, str);
}

static void *xcpc_excpt_memdup(xcpc_ctx ctx, void const *data, int size) {
	void *ddata;

	ddata = xcpc__malloc(xcpc_context_exhandler(ctx), NULL, size);
	memcpy(ddata, data, size);

	return ddata;
}

XCAPI void *xcpc__malloc(xcpc_ctx ctx, xcpc_res *pres, long size) {
	xcpc_res res;
	void *data;

	if ((data = malloc(size)) == NULL)
		XCPC_THROW(ctx, XCPCE_MEMORY, NULL);
	res = xcpc_resource_add(ctx, data, free);
	if (pres != NULL)
		*pres = res;

	return data;
}

XCAPI void *xcpc__realloc(xcpc_res res, long size) {
	void *data;

	if ((data = realloc(xcpc_resource_data(res), size)) == NULL)
		XCPC_THROW(xcpc_resource_ctx(res), XCPCE_MEMORY, NULL);
	xcpc_resource_set(res, 0, data, free);

	return data;
}

XCAPI char *xcpc__strdup(xcpc_ctx ctx, xcpc_res *pres, char const *str) {
	int len = strlen(str) + 1;
	char *dstr;

	dstr = xcpc__malloc(ctx, pres, len);
	memcpy(dstr, str, len);

	return dstr;
}

#if defined(HAVE_FCLOSE)

static void xcpc_hlp__fclose(void *data) {

	fclose((FILE *) data);
}

#if defined(HAVE_FOPEN)

XCAPI FILE *xcpc__fopen(xcpc_ctx ctx, xcpc_res *pres, char const *path,
			char const *mode) {
	xcpc_res res;
	FILE *file;

	if ((file = fopen(path, mode)) == NULL)
		XCPC_THROW(ctx, XCPCE_FOPEN, xcpc_excpt_strdup(ctx, path));
	res = xcpc_resource_add(ctx, file, xcpc_hlp__fclose);
	if (pres != NULL)
		*pres = res;

	return file;
}

#endif /* HAVE_FOPEN */

#if defined(HAVE_FDOPEN)

XCAPI FILE *xcpc__fdopen(xcpc_ctx ctx, xcpc_res *pres, int fd,
			 char const *mode) {
	xcpc_res res;
	FILE *file;

	if ((file = fdopen(fd, mode)) == NULL)
		XCPC_THROW(ctx, XCPCE_FDOPEN, (void *) (long) fd);
	res = xcpc_resource_add(ctx, file, xcpc_hlp__fclose);
	if (pres != NULL)
		*pres = res;

	return file;
}

#endif /* HAVE_FDOPEN */

#if defined(HAVE_FREOPEN)

XCAPI FILE *xcpc__freopen(xcpc_res res, char const *path,
			  char const *mode) {
	FILE *ofile;
	FILE *file;

	ofile = (FILE *) xcpc_resource_data(res);
	if ((file = freopen(path, mode, ofile)) == NULL)
		XCPC_THROW(xcpc_resource_ctx(res), XCPCE_FREOPEN,
			   xcpc_excpt_strdup(xcpc_resource_ctx(res), path));
	xcpc_resource_set(res, 0, file, xcpc_hlp__fclose);

	return file;
}

#endif /* HAVE_FREOPEN */

#endif /* HAVE_FCLOSE */

#if defined(HAVE_CLOSE)

static void xcpc_hlp__close(void *data) {

	close((int) (long) data);
}

#if defined(HAVE_OPEN)

XCAPI int xcpc__open(xcpc_ctx ctx, xcpc_res *pres, char const *path,
		     int flags, int mode) {
	int fd;
	xcpc_res res;

	if ((fd = open(path, flags, mode)) == -1)
		XCPC_THROW(ctx, XCPCE_OPEN, xcpc_excpt_strdup(ctx, path));
	res = xcpc_resource_add(ctx, (void *) (long) fd, xcpc_hlp__close);
	if (pres != NULL)
		*pres = res;

	return fd;
}

#endif /* HAVE_OPEN */

#endif /* HAVE_CLOSE */

#if defined(HAVE_REMOVE)

static void xcpc_hlp__remove(void *data) {

	remove((char *) data);
	free(data);
}

XCAPI void xcpc__remove(xcpc_ctx ctx, const char *path) {

	if (remove(path))
		XCPC_THROW(ctx, XCPCE_REMOVE, xcpc_excpt_strdup(ctx, path));
}

XCAPI void xcpc__push_remove(xcpc_ctx ctx, xcpc_res *pres, char const *path) {
	void *data;
	xcpc_res rres, res;

	data = xcpc__malloc(ctx, &rres, strlen(path));
	strcpy(data, path);
	res = xcpc_resource_add(ctx, data, xcpc_hlp__remove);
	xcpc_resource_del(rres, 0);
	if (pres != NULL)
		*pres = res;
}

#endif /* HAVE_REMOVE */

#if defined(HAVE_MUNMAP)

typedef struct s_xcpc_munmap_data {
	void *addr;
	long size;
} xcpc_munmap_data;


static void xcpc_hlp__munmap(void *data) {
	xcpc_munmap_data *udata = (xcpc_munmap_data *) data;

	munmap(udata->addr, udata->size);
	free(udata);
}

#if defined(HAVE_MMAP)

XCAPI void *xcpc__mmap(xcpc_ctx ctx, xcpc_res *pres, void *start, size_t length,
		       int prot, int flags, int fd, off_t offset) {
	xcpc_res rres, res;
	xcpc_munmap_data *udata;

	udata = xcpc__malloc(ctx, &rres, sizeof(xcpc_munmap_data));
	udata->size = length;
	if ((udata->addr = mmap(start, length, prot, flags, fd,
				offset)) == (void *) -1)
		XCPC_THROW(ctx, XCPCE_MMAP, NULL);
	res = xcpc_resource_add(ctx, udata, xcpc_hlp__munmap);
	xcpc_resource_del(rres, 0);
	if (pres != NULL)
		*pres = res;

	return udata->addr;
}

#endif /* HAVE_MMAP */

#endif /* HAVE_MUNMAP */

#if defined(HAVE_WRITE)

XCAPI ssize_t xcpc__write(xcpc_ctx ctx, int fd, const void *buf, size_t count) {
	ssize_t res;

	if ((res = write(fd, buf, count)) != count)
		XCPC_THROW(ctx, XCPCE_WRITE, (void *) (long) fd);

	return res;
}

#endif /* HAVE_WRITE */

#if defined(HAVE_READ)

XCAPI ssize_t xcpc__read(xcpc_ctx ctx, int fd, void *buf, size_t count) {
	ssize_t res;

	if ((res = read(fd, buf, count)) != count)
		XCPC_THROW(ctx, XCPCE_READ, (void *) (long) fd);

	return res;
}

#endif /* HAVE_READ */

#if defined(HAVE_FWRITE)

XCAPI size_t xcpc__fwrite(xcpc_ctx ctx, const void *ptr, size_t size, size_t nmemb,
			  FILE *stream) {
	size_t res;

	if ((res = fwrite(ptr, size, nmemb, stream)) != nmemb)
		XCPC_THROW(ctx, XCPCE_FWRITE, stream);

	return res;
}

#endif /* HAVE_FWRITE */

#if defined(HAVE_FREAD)

XCAPI size_t xcpc__fread(xcpc_ctx ctx, void *ptr, size_t size, size_t nmemb,
			 FILE *stream) {
	size_t res;

	if ((res = fread(ptr, size, nmemb, stream)) != nmemb)
		XCPC_THROW(ctx, XCPCE_FREAD, stream);

	return res;
}

#endif /* HAVE_FREAD */

#if defined(HAVE_STAT)

XCAPI void xcpc__stat(xcpc_ctx ctx, const char *path, struct stat *buf) {

	if (stat(path, buf))
		XCPC_THROW(ctx, XCPCE_STAT, xcpc_excpt_strdup(ctx, path));
}

#endif /* HAVE_STAT */

#if defined(HAVE_FSTAT)

XCAPI void xcpc__fstat(xcpc_ctx ctx, int fd, struct stat *buf) {

	if (fstat(fd, buf))
		XCPC_THROW(ctx, XCPCE_FSTAT, (void *) (long) fd);
}

#endif /* HAVE_FSTAT */

#if defined(HAVE_MKDIR)

XCAPI void xcpc__mkdir(xcpc_ctx ctx, const char *path, mode_t mode) {

	if (mkdir(path, mode))
		XCPC_THROW(ctx, XCPCE_MKDIR, xcpc_excpt_strdup(ctx, path));
}

#endif /* HAVE_MKDIR */

#if defined(HAVE_RMDIR)

XCAPI void xcpc__rmdir(xcpc_ctx ctx, const char *path) {

	if (rmdir(path))
		XCPC_THROW(ctx, XCPCE_RMDIR, xcpc_excpt_strdup(ctx, path));
}

#endif /* HAVE_RMDIR */

#if defined(HAVE_OPENDIR)

XCAPI void *xcpc__opendir(xcpc_ctx ctx, const char *path) {
	DIR *dir;

	if ((dir = opendir(path)) == NULL)
		XCPC_THROW(ctx, XCPCE_OPENDIR, xcpc_excpt_strdup(ctx, path));

	return dir;
}

#endif /* HAVE_OPENDIR */

#if defined(HAVE_PIPE)

XCAPI void xcpc__pipe(xcpc_ctx ctx, int *fds) {

	if (pipe(fds))
		XCPC_THROW(ctx, XCPCE_PIPE, NULL);
}

#endif /* HAVE_PIPE */


#if defined(HAVE_SOCKET)

XCAPI int xcpc__socket(xcpc_ctx ctx, int domain, int type, int protocol) {
	int sfd;

	if ((sfd = socket(domain, type, protocol)) == -1)
		XCPC_THROW(ctx, XCPCE_SOCKET, NULL);

	return sfd;
}

#endif /* HAVE_SOCKET */

#if defined(HAVE_BIND)

XCAPI void xcpc__bind(xcpc_ctx ctx, int sfd, const struct sockaddr *addr,
		      int alen) {

	if (bind(sfd, addr, alen))
		XCPC_THROW(ctx, XCPCE_BIND, xcpc_excpt_memdup(ctx, addr, alen));
}

#endif /* HAVE_BIND */

#if defined(HAVE_CONNECT)

XCAPI void xcpc__connect(xcpc_ctx ctx, int sfd, const struct sockaddr *addr,
			 int alen) {

	if (connect(sfd, addr, alen))
		XCPC_THROW(ctx, XCPCE_CONNECT, xcpc_excpt_memdup(ctx, addr, alen));
}

#endif /* HAVE_CONNECT */

#if defined(HAVE_TRUNCATE)

XCAPI void xcpc__truncate(xcpc_ctx ctx, char const *path, off_t len) {

	if (truncate(path, len))
		XCPC_THROW(ctx, XCPCE_TRUNCATE, xcpc_excpt_strdup(ctx, path));
}

#endif /* HAVE_TRUNCATE */

#if defined(HAVE_FTRUNCATE)

XCAPI void xcpc__ftruncate(xcpc_ctx ctx, int fd, off_t len) {

	if (ftruncate(fd, len))
		XCPC_THROW(ctx, XCPCE_FTRUNCATE, (void *) (long) fd);
}

#endif /* HAVE_FTRUNCATE */

#if defined(HAVE_LSEEK)

XCAPI off_t xcpc__lseek(xcpc_ctx ctx, int fd, off_t off, int whence) {
	off_t boff;

	if ((boff = lseek(fd, off, whence)) == -1)
		XCPC_THROW(ctx, XCPCE_LSEEK, (void *) (long) fd);

	return boff;
}

#endif /* HAVE_LSEEK */

